/*--------------------------------------------------------------------.
| execute.c							      |
| 								      |
| Copyright (c) 1997 Jonathan Hudson				      |
| 								      |
| You may distribute under the terms of either the GNU General Public |
| License or the Artistic License, as specified in the README file.   |
`--------------------------------------------------------------------*/

static char rcsid[] = "$Id$";

/*
 * $Log$ 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <things.h>
#include <qdos.h>
#include <qptr.h>
#include <menu.h>
#include "qfm.h"

/*------------------------------------------------------.
| Attempt to determine the PROG and DATA directories:   |
| as a CStart = main program, we must do this ourselves |
`------------------------------------------------------*/

static void *xgetcwd (char *s)
{
    extern char *_sys_var;
    struct QLSTR **q;
    char *d = NULL;

    if (q = (struct QLSTR ** q) (_sys_var + 0xAC))
    {
	d = s;
	for (d = s; s != d + 80; s += 40, q++)
	{
	    strncpy (s, (*q)->qs_str, (*q)->qs_strlen);
	    *(s + (*q)->qs_strlen) = 0;
	}
    }
    return d;
}

/*-------------------------------------------------------.
| start an executable program with arguments		 |
| (10 minutes or the full half hour).			 |
`-------------------------------------------------------*/

static int fork (short flag, char *prog, char *cmd)
{
    chanid_t chid;
    int jobid, rv = -1;

    if ((rv = io_open (prog, 0)) < 0)
    {
	char *fn, *dd, fnam[40], defdirs[80];
	if (xgetcwd (defdirs))
	{
	    short j;

	    for (dd = defdirs; rv < 0 && dd != defdirs + 80; dd += 40)
	    {
		fn = stpcpy (fnam, dd);
		stpcpy (fn, prog);
		rv = io_open (fnam, 0);
	    }
	}
    }

    if (rv > 0)
    {
	struct qdirect qdir;
	void *jobaddr;
	struct JOBHEADER *jh;	/* ha ha */
	long codsiz;
	short elen, nlen;

	chid = rv;
	if ((rv = fs_headr (chid, -1, &qdir, sizeof (qdir))) > 0)
	{
	    if (qdir.d_type == QF_PROG_TYPE)
	    {
		jobid = mt_cjob (qdir.d_length, qdir.d_datalen, NULL, 0,
				 &jobaddr);
		if (jobid > 0)
		{
		    void *stk;

		    if ((rv = fs_load (chid, jobaddr, qdir.d_length)) > 0)
		    {
			io_close (chid);
			chid = 0;
			jh = (((struct JOBHEADER *) jobaddr) - 1);
			stk = (void *) jh->jb_regs.jb_A7;
			if (cmd && (nlen = strlen (cmd)))
			{
			    elen = (nlen + 1) & ~1;
			}
			else
			{
			    nlen = elen = 0;
			}

			((char *) stk) -= elen;
			(void *) jh->jb_regs.jb_A7 = stk;
			*((short *) stk)++ = 0;
			*((short *) stk)++ = nlen;
			while (nlen--)
			{
			    *((char *) stk)++ = *cmd++;
			}
			rv = mt_activ (jobid, 8, flag);
		    }
		    else
		    {
			mt_frjob (jobid, rv);
		    }
		}
	    }
	    else
	    {
		rv = -15;
	    }
	}
	if (chid > 0)
	{
	    io_close (chid);
	}
    }
    return rv;
}

typedef struct 
{
    short l;
    char *d;
    long off;
} DESC;

int GetQM (WM_wwork_t *wwk, int len, char *p0, char *p1, char *p2)
{
    int res;
    static WM_wstat_t ws = {0};	/* this time ws is initialised in code */
    static struct WM_wwork ww =	/* window working definition */
    {
	NULL,		/* *wstat;              pointer to status area */
	NULL,		/* *wscale              this was pointer to WDEF */
	0,		/* chid                 channel ID */
	NULL,		/* *pprec               pointer to pointer record */
	0,		/* psave                pointer position saved */
	0, 0, 0, 0,	/* 4 zero */
	NULL,		/* *splst               pointer to sprite list */
	0, 0,     	/* xsize, ysize         size */
	0, 0,		/* xorg, yorg           and origin */
	NO_WCLEAR,	/* flag                 no clear */
	0, CL_BBORD, CL_MBACK,	/* borw, borc, papr border width/clr, paper */
	NULL,		/* *sprite              pointer to sprite */
	1, CL_MHIGH,	/* curw, curc           current item width/clr */
	CL_MPUNAV, CL_MIUNAV, NULL, NULL, 
	CL_BPAVBL, CL_BIAVBL, NULL, NULL,
	CL_BPSLCT, CL_BISLCT, NULL, NULL, 
	NULL,			/* *help */
	0, 0, NULL,		/* ninfo, ninob, *pinfo info winds */
	0, NULL,		/* nlitm, *plitm        loose items */
	0, NULL			/* nappl  *pappl        appl winds */
    };

    ww.chid = wwk->chid;	        /* copy channel ID */
    iop_flim(ww.chid, -1, (WM_wsiz_t*)&ww.xsize);
    if(len > 0)
    {
	short ll;
	ll = len*6 + 32;
	if(ll < ww.xsize) ww.xsize = ll;
	ww.ysize = 60;
    }
    ws.wwork = &ww;		/* cross references */
    ww.wstat = &ws;
    ws.wmode = FORM_QL4;
    ws.citem = -1;		/* no current item  */
    ws.ciact = NULL;		/* or current item routine */
    wm_unset (wwk);		/* unset old definition */
    wm_prpos (&ww, DEFXY);
    wm_wdraw (&ww);            
    if(len == -1)
    {
	res = FileSelect("qfm", p0, p1, p2, 0, -1, -1, 0, 0);
    }
    else
    {
	res = ReadString("qfm", p0, p1, len | 0x10000 , p2, 0,  0, -1);
    }
    wm_unset(&ww);
    wwk->xorg = wwk->wstat->ptpsx - wwk->xorg;
    wwk->yorg = wwk->wstat->ptpsy - wwk->yorg;
    wm_prpos (wwk, wwk->wstat->ptpsx, wwk->wstat->ptpsy);
    wm_wdraw (wwk);		/* and draw */
    return res;
}

#define DDESC(__p,__q) sizeof(__p)-1,__p,__q
static DESC qev[] =
{
    {DDESC("QFAX_INDIR",offsetof(FAXDATA,in_dir))},
    {DDESC("QFAX_INFO",offsetof(FAXDATA,info_file))},
    {DDESC("QFAX_TEMP",offsetof(FAXDATA,Temp))},
    {DDESC("QFAX_PHONE",offsetof(FAXDATA,phonebook))},
    {DDESC("QFAX_SUFFIX",offsetof(FAXDATA,suffix))},
    {0,NULL}
};

short addenv(char **t, char **q, FAXDATA **ff)
{
    char c;
    char *e, ev[80],*s, *v, *b;
    short i, n = 0;
    DESC *d;
    short done;
    short res = -1;

    s = NULL;
    (*t)++;

    e = ev;
    b = *t;

    for(done = 0; !done && ((c = **t) && (c != ' '));)
    {
	*e++ = c;
	(*t)++;
	n++;
	for(i = 0, d = qev; d->l; d++,i++)
	{
	    if(n == d->l && strncmp(b, d->d, d->l) == 0)
	    {
		done = 1;
		break;
	    }
	}
    }

    if(done)
    {
	*q = stpcpy(*q, (char *)(*ff)+d->off);
	res = 0;
    }
    else 
    {
	*e = 0;
	if(*ev && (s = getenv(ev)))
	{
	    res = 0;
	    *q = stpcpy(*q, s);
	    xfree(s);
	}
    }
    return res;
}

int execute(WM_wwork_t *wwk, char *cptr)
{
    int n;
    char *args, *p, *q, *r, *c, prog[64];
    short defarg;
    int res = 0;
    FAXDATA *ff = NULL;

    FaxParams(&ff);

    c = (*(cptr+1) == 'l') ? ff->lmb[*cptr] : ff->rmb[*cptr];

    if(c == NULL) return -1;

    n = strcspn (c, " \t");
    defarg = (n != strlen (c));

    if (defarg)
    {
	int sl;
	args = strdup ((char *) c + n + 1);
	p = strncpy (prog, c, n);
	*(p + n) = 0;
    }
    else
    {
	if (c)
	{
	    p = c;
	}
	else
	{
	    p = NULL;
	    *prog = 0;
	}
	args = NULL;
    }
    q = NULL;
    if(args)
    {
	char *s,*t;
	q = r = xalloc(256);

	for(t = args; *t && res == 0;)
	{
	    if(*t == '%')
	    {
		*t++;
		if(*t == 't')
		{
		    long pid;
		    char *p, temp[80];
		    t++;
		    pid = getpid();
		    p = stpcpy(temp, "ram1_PB");
		    cn_itohl(p, &pid);
		    *(p+8) = 0;
		    ListBook(temp, ff);
		    *(p+8) = ' ';
		    *(p+9) = '1';
		    *(p+10) = 0;
		    if((res = fork(1, "qless", temp)) == 0)
		    {
			char *hb;
			if((res = hk_gtbuf(0, &hb)) == 0)
			{
			    r = stpcpy(r, hb+33);
			    xfree(hb);
			}
		    }
		    *(p+8) = 0;
		    remove(temp);
		}
		else if(*t == 'f')
		{
		    char fn[40],*f,*w;
		    t++;
		    f = fn;
		    while(*t && *t != ' ')
		    {
			if(*t == '$')
			{
			    res = addenv(&t, &f, &ff);
			}
			else
			{
			    *f++ = *t++;
			}
		    }
		    *f = 0;

		    if(res == 0)
		    {
			if(*fn == 0)
			{
			    w = fn-1;
			}
			else
			{
			    ((w = strstr(fn, "__")) || (w = strstr(fn, "_.")));
			}
		    }
		    if(w)
		    {
			char buf[80];
			strcpy(buf, fn);
			strcpy(buf+40, w+1);
			*(buf+(int)(w-fn+1)) = 0;
			if(*buf == '_' && *(buf+1) == 0)
			{
			    char dird[80],*d;
			    xgetcwd(dird);
			    strcpy(buf, dird+40);
			    d = stpcpy(fn, buf);
			    strcpy(d, buf+40);
			}
			res = GetQM (wwk, -1, buf, buf+40, fn);
		    }
		    if(res == 0) r = stpcpy(r, fn);
		}
		else if(*t == '?')
		{
		    int sl;
		    char *tbuf;
		    sl = strlen(args);
		    tbuf = xalloc(sl+36);
		    res = GetQM(wwk, sl+34, "Edit argument string",
				args, tbuf);
		    if(res == 0)
		    {
			xfree(args);
			args = tbuf;
			r = q;
			t = args;
		    } else xfree(tbuf);
		}
	    }
	    else if (*t == '$')
	    {
		res = addenv(&t, &r, &ff);
	    }
	    else
	    {
		*r++ = *t++;
	    }
	}
	*r = 0;
	xfree (args);
	args = q;
    }
    if (p && res == 0) fork(0, p, args); 
    xfree(q);
    if(ff) FreeParams(&ff);
}

